@@ -32,7 +32,7 @@
 #
 # ==========================================================================
 
-version="16.6"
+version="19.0"
 
 # initialize common flags
 
@@ -44,6 +44,11 @@ api_key=""
 immediate=false
 express=false
 
+# undocumented -quick argument blocks isPreview
+# flags that work around PubMed SOLR preview server limits
+quick=false
+quickx=false
+
 email=""
 emailr=""
 emailx=""
@@ -65,6 +70,9 @@ label=""
 labels=""
 labelx=""
 
+verbose=false
+tranquil=false
+
 basx=""
 base="https://eutils.ncbi.nlm.nih.gov/entrez/eutils/"
 
@@ -93,6 +101,7 @@ empty=false
 stp=0
 
 rest=""
+qury=""
 
 # set up colors for error report
 
@@ -193,22 +202,28 @@ ParseStdin() {
   then
     mssg=$( cat )
     ParseMessage "$mssg" ENTREZ_DIRECT \
-                  dbase Db web_env WebEnv qry_key QueryKey \
+                  dbase Db web_env WebEnv qry_key QueryKey qury Query \
                   num Count stp Step toolx Tool emailx Email \
-                  labelx Labels debugx Debug logx Log timerx Elapsed
+                  labelx Labels quickx Quick debugx Debug logx Log timerx Elapsed
     if [ "$?" = 2 ]
     then
       # if no ENTREZ_DIRECT message present, support passing raw UIDs via stdin
       rest="$mssg"
     else
-      # future support for UIDs instantiated within message in lieu of Entrez History
-      rest=$( echo "$mssg" | xtract -pattern ENTREZ_DIRECT -sep "\n" -element Id )
+      # support for UIDs instantiated within message in lieu of Entrez History
+      rest=$( echo "$mssg" |
+              xtract -pattern ENTREZ_DIRECT -sep "\n" -element Id |
+              grep '.' | sort -n | uniq )
       if [ -z "$stp" ]
       then
         stp=1
       fi
       # hierarchy is -email argument, then Email XML field, then calculated email
       AdjustEmailAndTool
+      if [ "$quickx" = "Y" ]
+      then
+        quick=true
+      fi
       if [ "$debugx" = "Y" ]
       then
         debug=true
@@ -361,6 +376,28 @@ ParseCommonArgs() {
           exit 1
         fi
         ;;
+      -quick )
+        argsConsumed=$(($argsConsumed + 1))
+        shift
+        if [ $# -gt 0 ]
+        then
+          if [ "$1" = "true" ]
+          then
+            argsConsumed=$(($argsConsumed + 1))
+            shift
+            quick=true
+          elif [ "$1" = "false" ]
+          then
+            argsConsumed=$(($argsConsumed + 1))
+            shift
+            quick=false
+          else
+            quick=true
+          fi
+        else
+          quick=true
+        fi
+        ;;
       -debug )
         argsConsumed=$(($argsConsumed + 1))
         shift
@@ -383,7 +420,34 @@ ParseCommonArgs() {
           debug=true
         fi
         ;;
-      -log )
+       -verbose )
+        argsConsumed=$(($argsConsumed + 1))
+        shift
+        if [ $# -gt 0 ]
+        then
+          if [ "$1" = "true" ]
+          then
+            argsConsumed=$(($argsConsumed + 1))
+            shift
+            verbose=true
+          elif [ "$1" = "false" ]
+          then
+            argsConsumed=$(($argsConsumed + 1))
+            shift
+            verbose=false
+          else
+            verbose=true
+          fi
+        else
+          verbose=true
+        fi
+        ;;
+      -tranquil )
+        tranquil=true
+        argsConsumed=$(($argsConsumed + 1))
+        shift
+        ;;
+     -log )
         argsConsumed=$(($argsConsumed + 1))
         shift
         if [ $# -gt 0 ]
@@ -584,7 +648,17 @@ ErrorTail() {
 
   printf "${INIT}" >&2
   # display reformatted result in red letters
-  printf "${RED}${msg}${INIT}\n" >&2
+  lin=$( echo "${msg}" | wc -l)
+  if [ -n "$lin" ] && [ "$lin" -gt 25 ]
+  then
+    hd=$( echo "${msg}" | head -n 10 )
+    tl=$( echo "${msg}" | tail -n 10 )
+    printf "${RED}${hd}${INIT}\n" >&2
+    printf "${RED}...${INIT}\n" >&2
+    printf "${RED}${tl}${INIT}\n" >&2
+  else
+    printf "${RED}${msg}${INIT}\n" >&2
+  fi
   if [ "$goOn" = true ]
   then
     printf "${BLUE}${whc} ATTEMPT" >&2
@@ -699,6 +773,52 @@ RequestWithRetry() {
             # retry query
             res=$( "$@" )
             ;;
+          *\<WarningList\>* )
+            case "$res" in
+              *"<OutputMessage>No items found"* )
+                if [ "$tranquil" = true ]
+                then
+                  # -tranquil flag conditionally ignores no items found message
+                  goOn=false
+                else
+                  ref=$( echo "$res" | transmute -format indent -doctype "" )
+                  # question mark prints names of heterogeneous child objects
+                  errs=$( echo "$res" | xtract -pattern "WarningList/*" -element "?" )
+                  ErrorHead "$warn" "$when"
+                  PrintQuery "$@"
+                  if [ "$goOn" = true ]
+                  then
+                    # reconstruct indented ErrorList XML
+                    ref=$( echo "$res" | xtract -head "<WarningList>" -tail "<WarningList>" \
+                           -pattern "WarningList/*" -pfx "  " -element "*" )
+                  fi
+                  ErrorTail "$ref" "$whch"
+                  sleep 1
+                  when=$( date )
+                  # retry query
+                  res=$( "$@" )
+                fi
+                ;;
+              * )
+                ref=$( echo "$res" | transmute -format indent -doctype "" )
+                # question mark prints names of heterogeneous child objects
+                errs=$( echo "$res" | xtract -pattern "WarningList/*" -element "?" )
+                ErrorHead "$warn" "$when"
+                PrintQuery "$@"
+                if [ "$goOn" = true ]
+                then
+                  # reconstruct indented ErrorList XML
+                  ref=$( echo "$res" | xtract -head "<WarningList>" -tail "<WarningList>" \
+                         -pattern "WarningList/*" -pfx "  " -element "*" )
+                fi
+                ErrorTail "$ref" "$whch"
+                sleep 1
+                when=$( date )
+                # retry query
+                res=$( "$@" )
+                ;;
+            esac
+            ;;
           * )
             # success - no error message detected
             goOn=false
@@ -780,6 +900,120 @@ RunWithCommonArgs() {
   RunWithLogging "$@"
 }
 
+# break PubMed SOLR server requests into chunks
+
+# time esearch -db pubmed -query "PNAS [JOUR] AND 2020:2022 [PDAT]" | efetch -format uid -verbose > /dev/null
+
+SearchSOLR() {
+
+  local fr
+  local to
+  local md
+  local nx
+
+  qy="$1"
+  fr="$2"
+  to="$3"
+
+  if [ "$verbose" = false ]
+  then
+    printf "." >&2
+  fi
+
+  # -tranquil flag specifically suppresses "No items found" message and automatic retries
+  count=$( esearch -db pubmed -count "( ${qy} AND ${fr}:${to} [UID])" -tranquil < /dev/null )
+  if [ -n "$count" ]
+  then
+    if [ "$count" -eq 0 ]
+    then
+      if [ "$verbose" = true ]
+      then
+        printf "  %08d : %08d [UID] . %d\n" "$fr" "$to" "$count" >&2
+      fi
+    elif [ "$count" -le 9999 ]
+    then
+      if [ "$verbose" = true ]
+      then
+        printf "  %08d : %08d [UID] * %d\n" "$fr" "$to" "$count" >&2
+      fi
+      esearch -db pubmed -subset "( ${qy} AND ${fr}:${to} [UID])" -tranquil < /dev/null |
+      xtract -pattern eSearchResult -block IdList -sep "\n" -element Id |
+      sort -n | uniq
+    else
+      if [ "$verbose" = true ]
+      then
+        printf "  %08d : %08d [UID] - %d\n" "$fr" "$to" "$count" >&2
+      fi
+      # partition half way between the points
+      md=$(( fr / 2 + to / 2 ))
+      nx=$(( md + 1 ))
+      SearchSOLR "$qy" "$fr" "$md"
+      SearchSOLR "$qy" "$nx" "$to"
+    fi
+  else
+    if [ "$verbose" = true ]
+    then
+      printf "  %08d : %08d [UID] ~ 0\n" "$fr" "$to" >&2
+    fi
+  fi
+}
+
+GenerateFromSOLR() {
+
+  currentYear=$( date +"%Y" )
+  total="$num"
+
+  if [ "$total" -lt 1 ]
+  then
+    exit 0
+  fi
+
+  # comment out this conditional section to test with small samples
+  if [ "$total" -lt 10000 ]
+  then
+    esearch -db pubmed -uids "$qury" -tranquil < /dev/null
+    exit 0
+  fi
+
+  minPMID=1
+
+  # remove pubmed books inserted at front of list, but still retrieve 1000 PMIDs and take maximum
+  maxPMID=$(
+    esearch -db pubmed -subset "all [SB] NOT pubmed books [SB]" -retmax 1000 -tranquil < /dev/null |
+    xtract -pattern eSearchResult -max Id
+  )
+
+  # reality check for successful creation of maxPMID variable
+  if [ ! -n "$maxPMID" ]
+  then
+    echo "Unable to get top PMID" >&2
+    exit 0
+  fi
+
+  # occasional older records, newly found and submitted by publishers,
+  # will be lower in the sort order (EDAT is set to its PDAT value), so
+  # increment by 1000 to accommodate this rare situation
+  maxPMID=$(( maxPMID + 1000 ))
+
+  uids=$( SearchSOLR "$qury" "$minPMID" "$maxPMID" | sort -n | uniq )
+
+  if [ "$verbose" = false ]
+  then
+    printf "\n" >&2
+  fi
+
+  if [ -n "$uids" ]
+  then
+    cumulative=$( echo "$uids" | wc -l | tr -d ' ' )
+    if [ "$verbose" = true ]
+    then
+      echo "cumulative: ${cumulative},  total: ${total}" >&2
+    fi
+  fi
+
+  echo "$uids"
+}
+
 # break Entrez history server requests into chunks
 
 GenerateHistoryChunks() {
@@ -815,24 +1049,44 @@ GenerateHistoryChunks() {
 
 GenerateUidList() {
 
+  dbsx="$1"
+
+  fromSOLR=false
   if [ "$needHistory" = true ]
   then
-    dbsx="$1"
     chunk=25000
+    mxx=0
     if [ "$dbsx" = "pubmed" ]
     then
       chunk=10000
+      if [ "$quick" = false ]
+      then
+        mxx=9999
+      fi
     fi
-    GenerateHistoryChunks "$chunk" 0 0 |
+    GenerateHistoryChunks "$chunk" 0 "$mxx" |
     while read fr chnk
     do
-      RunWithCommonArgs nquire -get "$base" efetch.fcgi \
+      RunWithCommonArgs nquire -url "$base" efetch.fcgi \
         -query_key "$qry_key" -WebEnv "$web_env" -retstart "$fr" -retmax "$chnk" \
         -db "$dbsx" -rettype uilist -retmode text
     done
   else
     # otherwise obtain raw UIDs from available sources
-    if [ -n "$ids" ]
+    if [ -n "$qury" ] && [ "$dbsx" = "pubmed" ] && [ "$isPreview" = true ]
+    then
+      fromSOLR=true
+      slr=$( GenerateFromSOLR )
+      if [ -n "$slr" ]
+      then
+        nm=$( echo "$slr" | wc -l  | tr -d ' ' )
+        if [ -n "$nm" ] && [ -n "$num" ] && [ "$nm" -ne "$num" ]
+        then
+          echo "${INVT} WARNING: ${LOUD} Actual PMID count ${nm} does not match expected total ${num}${INIT}" >&2
+        fi
+      fi
+      echo "$slr"
+    elif [ -n "$ids" ]
     then
       echo "$ids"
     elif [ -n "$rest" ]
@@ -846,7 +1100,8 @@ GenerateUidList() {
     sed 's/[^a-zA-Z0-9_.]/ /g; s/^ *//' |
     fmt -w 1 |
     grep '.'
-  fi
+  fi |
+  sort -n | uniq
 }
 
 # special case accession to UID lookup functions
@@ -1020,7 +1275,7 @@ RunAccnSearch() {
 
   while read qry
   do
-    nquire -get "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi" \
+    nquire -url "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi" \
       -db "$dbase" -term "$qry" -retmax "$1" < /dev/null |
     xtract -pattern eSearchResult -sep "\n" -element IdList/Id
   done
@@ -1267,6 +1522,10 @@ WriteEDirect() {
     echo "  </Labels>"
   fi
 
+  if [ "$quick" = true ] || [ "$quickx" = "Y" ]
+  then
+    echo "  <Quick>Y</Quick>"
+  fi
   if [ "$debug" = true ] || [ "$debugx" = "Y" ]
   then
     echo "  <Debug>Y</Debug>"
