@@ -38,7 +38,7 @@ import (
 	"hash/crc32"
 	"io"
 	"os"
-	"path"
+	"path/filepath"
 	"runtime"
 	"strconv"
 	"strings"
@@ -46,6 +46,181 @@ import (
 	"time"
 )
 
+// fetchOneXMLRecord is an internal function for fetching an XML record from the archive
+func fetchOneXMLRecord(id, base, pfx, sfx string, zipp bool, buf bytes.Buffer) string {
+
+	if id == "" {
+		return ""
+	}
+
+	id = strings.TrimPrefix(id, "PMC")
+
+	dir, file := ArchiveTrie(id)
+
+	if dir == "" || file == "" {
+		return ""
+	}
+
+	if zipp {
+		sfx += ".gz"
+	}
+
+	fpath := filepath.Join(base, dir, pfx+file+sfx)
+	if fpath == "" {
+		return ""
+	}
+
+	iszip := zipp
+
+	inFile, err := os.Open(fpath)
+
+	// if failed to find ".xml" or ".e2x" file, try ".xml.gz" or ".e2x.gz" without requiring -gzip
+	if err != nil && os.IsNotExist(err) && !zipp {
+		iszip = true
+		fpath := filepath.Join(base, dir, pfx+file+sfx+".gz")
+		if fpath == "" {
+			return ""
+		}
+		inFile, err = os.Open(fpath)
+	}
+	if err != nil {
+		msg := err.Error()
+		if !strings.HasSuffix(msg, "no such file or directory") && !strings.HasSuffix(msg, "cannot find the path specified.") {
+			fmt.Fprintf(os.Stderr, "%s\n", msg)
+		}
+		return ""
+	}
+
+	defer inFile.Close()
+
+	brd := bufio.NewReader(inFile)
+
+	if iszip {
+
+		zpr, err := gzip.NewReader(brd)
+
+		defer zpr.Close()
+
+		if err == nil {
+			// copy and decompress cached file contents
+			buf.ReadFrom(zpr)
+		}
+
+	} else {
+
+		// copy cached file contents
+		buf.ReadFrom(brd)
+	}
+
+	str := buf.String()
+
+	return str
+}
+
+// FetchPubMedRecord returns the PubmedArticle XML for a single PMID
+func FetchPubMedRecord(id string) string {
+
+	if id == "" {
+		return ""
+	}
+
+	var buf bytes.Buffer
+
+	// obtain path from environment variable
+	base := os.Getenv("EDIRECT_PUBMED_MASTER")
+	if base != "" {
+		if !strings.HasSuffix(base, "/") {
+			base += "/"
+		}
+	}
+
+	archiveBase := base + "Archive"
+
+	// check to make sure local archive is mounted
+	_, err := os.Stat(archiveBase)
+	if err != nil && os.IsNotExist(err) {
+		fmt.Fprintf(os.Stderr, "\nERROR: Local archive and search index is not mounted\n\n")
+		os.Exit(1)
+	}
+
+	str := fetchOneXMLRecord(id, archiveBase, "", ".xml", true, buf)
+
+	// trim header now included in archive XML files
+	if str != "" {
+		pos := strings.Index(str, "<PubmedArticle>")
+		if pos > 0 {
+			// remove leading xml and DOCTYPE lines
+			str = str[pos:]
+		}
+	}
+
+	return str
+}
+
+// FetchPMCRecord returns the PMCExtract XML for a single PMCID
+func FetchPMCRecord(id string) string {
+
+	if id == "" {
+		return ""
+	}
+
+	id = strings.TrimPrefix(id, "PMC")
+
+	var buf bytes.Buffer
+
+	// obtain path from environment variable
+	base := os.Getenv("EDIRECT_PMC_MASTER")
+	if base != "" {
+		if !strings.HasSuffix(base, "/") {
+			base += "/"
+		}
+	}
+
+	archiveBase := base + "Archive"
+
+	// check to make sure local archive is mounted
+	_, err := os.Stat(archiveBase)
+	if err != nil && os.IsNotExist(err) {
+		fmt.Fprintf(os.Stderr, "\nERROR: Local archive and search index is not mounted\n\n")
+		os.Exit(1)
+	}
+
+	str := fetchOneXMLRecord(id, archiveBase, "PMC", ".xml", true, buf)
+
+	// trim header now included in archive XML files
+	if str != "" {
+		pos := strings.Index(str, "<PMCExtract>")
+		if pos > 0 {
+			// remove any leading xml and DOCTYPE lines
+			str = str[pos:]
+		}
+	}
+
+	return str
+}
+
+// GzipString allows separate compression of xml + DOCTYPE header and PubmedArticle record
+func GzipString(str string) []byte {
+
+	if str == "" {
+		return nil
+	}
+
+	var buf bytes.Buffer
+
+	// for small files, use regular gzip
+	gzWriter, err := gzip.NewWriterLevel(&buf, gzip.DefaultCompression)
+	if err != nil {
+		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
+		return nil
+	}
+
+	gzWriter.Write([]byte(str))
+	gzWriter.Close()
+
+	return buf.Bytes()
+}
+
 // CreateUIDReader sends PMIDs and their numeric orders down a channel.
 // This allows detection of updated records that appear shortly after
 // earlier versions, preventing the wrong version from being saved.
@@ -92,6 +267,48 @@ func CreateUIDReader(in io.Reader) <-chan XMLRecord {
 	return out
 }
 
+// ReadsUIDsFromString reads a comma-separated string of uids and sends through channel
+func ReadsUIDsFromString(uids string) <-chan XMLRecord {
+
+	if uids == "" {
+		return nil
+	}
+
+	out := make(chan XMLRecord, ChanDepth())
+	if out == nil {
+		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create uid string reader channel\n")
+		os.Exit(1)
+	}
+
+	// uidStringReader reads uids from input stream and sends through channel
+	uidStringReader := func(uids string, out chan<- XMLRecord) {
+
+		// close channel when all records have been processed
+		defer close(out)
+
+		items := strings.Split(uids, ",")
+
+		idx := 0
+		for _, item := range items {
+
+			idx++
+
+			pos := strings.Index(item, ".")
+			if pos >= 0 {
+				// remove version suffix
+				item = item[:pos]
+			}
+
+			out <- XMLRecord{Index: idx, Text: item}
+		}
+	}
+
+	// launch single uid string reader goroutine
+	go uidStringReader(uids, out)
+
+	return out
+}
+
 func mapXMLtoASN(node *XMLNode, proc func(string)) {
 
 	if node == nil || proc == nil {
@@ -220,9 +437,12 @@ func mapXMLtoASN(node *XMLNode, proc func(string)) {
 	doASNtree(node, 0, false)
 }
 
+// XMLDoctypeGzipLen is length of gzipped xml + DOCTYPE header, added as separate gzip packet in front of each compressed record
+const XMLDoctypeGzipLen = 183
+
 // CreateStashers saves records to archive, multithreaded for performance, use of UID
 // position index allows it to prevent earlier version from overwriting later version
-func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, report int, inp <-chan XMLRecord) <-chan string {
+func CreateStashers(stsh, parent, indx, pfx, sfx string, hash, zipp, asn, pmc bool, report int, inp <-chan XMLRecord) <-chan string {
 
 	if inp == nil {
 		return nil
@@ -266,7 +486,7 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 		idx, ok := inUse[id]
 
 		if ok {
-			if idx < index {
+			if index < idx {
 				// later version is being written by another goroutine, skip this
 				return BAIL
 			}
@@ -309,6 +529,37 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 		tlock.Unlock()
 	}
 
+	// common xml + DOCTYPE header for PubmedArticle XML
+	pmaSetHead := `<?xml version="1.0" encoding="UTF-8"?>
+<!DOCTYPE PubmedArticle PUBLIC "-//NLM//DTD PubMedArticle, 1st January 2019//EN" "https://dtd.nlm.nih.gov/ncbi/pubmed/out/pubmed_190101.dtd">
+`
+
+	// common xml without DOCTYPE header for PMCExtract XML (derived from BioC)
+	pmcSetHead := `<?xml version="1.0" encoding="UTF-8"?>
+`
+
+	xmlString := ""
+	if pmc {
+		xmlString = pmcSetHead
+	} else {
+		xmlString = pmaSetHead
+	}
+
+	// make gzip-compressed byte array of xml + DOCTYPE header
+	xmlDoctype := GzipString(xmlString)
+
+	if !pmc {
+
+		// reality check on expected length
+		pmaHeadLen := len(xmlDoctype)
+		// if it changes due to a new release date, or new gzip software,
+		// delete the Archive, remove Pubmed sentinel files, update constant
+		// in source code, and reindex
+		if pmaHeadLen != XMLDoctypeGzipLen {
+			fmt.Fprintf(os.Stdout, "Unexpected xmlDoctype length %d\n\n", pmaHeadLen)
+		}
+	}
+
 	// stashRecord saves individual XML record to archive file accessed by trie
 	stashRecord := func(str, id string, index int) string {
 
@@ -318,9 +569,9 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 			id = id[:pos]
 		}
 
-		var arry [132]rune
-		trie := MakeArchiveTrie(id, arry)
-		if trie == "" {
+		dir, file := ArchiveTrie(id)
+
+		if dir == "" || file == "" {
 			return ""
 		}
 
@@ -344,6 +595,16 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 				}
 			}
 			str = txt
+		} else if pmc {
+			// for XML, add trailing newline, needed for streaming without decompressing
+			if !strings.HasSuffix(str, "\n") {
+				str += "\n"
+			}
+		} else {
+			// for XML, add trailing newline, needed for streaming without decompressing
+			if !strings.HasSuffix(str, "\n") {
+				str += "\n"
+			}
 		}
 
 		attempts := 5
@@ -374,7 +635,7 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 		// delete lock after writing file
 		defer freeFile(id)
 
-		dpath := path.Join(stash, trie)
+		dpath := filepath.Join(stsh, dir)
 		if dpath == "" {
 			return ""
 		}
@@ -383,7 +644,7 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 			fmt.Fprintf(os.Stderr, "%s\n", err.Error())
 			return ""
 		}
-		fpath := path.Join(dpath, id+sfx)
+		fpath := filepath.Join(dpath, pfx+file+sfx)
 		if fpath == "" {
 			return ""
 		}
@@ -395,7 +656,7 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 			return ""
 		}
 
-		res := ""
+		res := id
 
 		if hash {
 			// calculate hash code for verification table
@@ -407,35 +668,13 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 
 		if zipp {
 
-			zpr, err := gzip.NewWriterLevel(fl, gzip.DefaultCompression)
-
-			if err == nil {
-
-				wrtr := bufio.NewWriter(zpr)
-
-				// compress and copy record to file
-				wrtr.WriteString(str)
-				if !strings.HasSuffix(str, "\n") {
-					wrtr.WriteString("\n")
-				}
-
-				err = wrtr.Flush()
-				if err != nil {
-					fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-					return ""
-				}
-
-				err = zpr.Close()
-				if err != nil {
-					fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-					return ""
-				}
-
-			} else {
-
-				fmt.Fprintf(os.Stderr, "%s\n", err.Error())
+			if !asn {
+				fl.Write(xmlDoctype)
 			}
 
+			zp := GzipString(str)
+			fl.Write(zp)
+
 		} else {
 
 			// copy uncompressed record to file
@@ -468,6 +707,11 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 
 			ext.Ident = FindIdentifier(ext.Text, parent, find)
 
+			// skip BioC records with explicit 'unknown' identifier
+			if ext.Ident == "unknown" {
+				continue
+			}
+
 			hsh := stashRecord(ext.Text, ext.Ident, ext.Index)
 
 			res := ext.Ident
@@ -501,84 +745,171 @@ func CreateStashers(stash, parent, indx, sfx string, hash, zipp, asn bool, repor
 	return out
 }
 
-// CreateFetchers returns uncompressed records from archive, multithreaded for speed
-func CreateFetchers(stash, sfx string, zipp bool, inp <-chan XMLRecord) <-chan XMLRecord {
+// CreateDeleter reads PMIDs, deletes them in the archive, and sends them
+// down a channel to have the affected inverted index cache files removed.
+func CreateDeleter(stsh string, in io.Reader) <-chan string {
+
+	if stsh == "" || in == nil {
+		return nil
+	}
+
+	out := make(chan string, ChanDepth())
+	if out == nil {
+		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create deleter channel\n")
+		os.Exit(1)
+	}
+
+	recordDeleter := func(in io.Reader, out chan<- string) {
+
+		// close channel when all records have been processed
+		defer close(out)
+
+		scanr := bufio.NewScanner(in)
+
+		verbose := false
+		// set verbose flag from environment variable
+		env := os.Getenv("EDIRECT_PUBMED_VERBOSE")
+		if env == "Y" || env == "y" {
+			verbose = true
+		}
+
+		for scanr.Scan() {
+
+			// read lines of identifiers
+			id := scanr.Text()
+
+			id = strings.TrimPrefix(id, "PMC")
+			id = strings.TrimSuffix(id, "\n")
+
+			pos := strings.Index(id, ".")
+			if pos >= 0 {
+				// remove version suffix
+				id = id[:pos]
+			}
+
+			dir, file := ArchiveTrie(id)
+
+			if dir == "" || file == "" {
+				continue
+			}
+
+			dpath := filepath.Join(stsh, dir, file+".xml.gz")
+			if dpath == "" {
+				continue
+			}
+
+			os.Remove(dpath)
+			if verbose {
+				fmt.Fprintf(os.Stderr, "DEL PMD %s\n", dpath)
+			}
+
+			out <- id
+		}
+	}
+
+	// launch single deleter goroutine
+	go recordDeleter(in, out)
+
+	return out
+}
+
+// CreateClearer clears incremental index files whose underlying records have changed
+func CreateClearer(indexBase, invertBase string, inp <-chan string) <-chan string {
 
 	if inp == nil {
 		return nil
 	}
 
+	out := make(chan string, ChanDepth())
+	if out == nil {
+		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create clearer channel\n")
+		os.Exit(1)
+	}
+
+	// maps to track index and inverted files that were deleted
+	deletedIdx := make(map[string]bool)
+
+	deletedInv := make(map[string]bool)
+
+	clearStaleIndices := func(inp <-chan string, out chan<- string) {
+
+		// close channel when all records have been processed
+		defer close(out)
+
+		verbose := false
+		// set verbose flag from environment variable
+		env := os.Getenv("EDIRECT_PUBMED_VERBOSE")
+		if env == "Y" || env == "y" {
+			verbose = true
+		}
+
+		for id := range inp {
+
+			id = strings.TrimPrefix(id, "PMC")
+			id = strings.TrimSuffix(id, "\n")
+
+			pos := strings.Index(id, ".")
+			if pos >= 0 {
+				// remove version from UID
+				id = id[:pos]
+			}
+
+			dir, idx := IndexTrie(id)
+			if dir == "" || idx == "" {
+				continue
+			}
+
+			dpath := filepath.Join(indexBase, dir, idx+".e2x.gz")
+			deletedIdx[dpath] = true
+
+			inv := InvertTrie(id)
+			if inv == "" {
+				continue
+			}
+
+			dpath = filepath.Join(invertBase, inv+".inv.gz")
+			deletedInv[dpath] = true
+
+			out <- id
+		}
+
+		// read (uniqued) maps and delete stale index and inverted files
+		for str := range deletedIdx {
+
+			os.Remove(str)
+			if verbose {
+				fmt.Fprintf(os.Stderr, "DEL IDX %s\n", str)
+			}
+		}
+
+		for str := range deletedInv {
+
+			os.Remove(str)
+			if verbose {
+				fmt.Fprintf(os.Stderr, "DEL INV %s\n", str)
+			}
+		}
+	}
+
+	// launch single clearer goroutine
+	go clearStaleIndices(inp, out)
+
+	return out
+}
+
+// CreateFetchers returns uncompressed records from archive, multithreaded for speed
+func CreateFetchers(stsh, db, pfx, sfx string, zipp bool, inp <-chan XMLRecord) <-chan XMLRecord {
+
+	if inp == nil || stsh == "" {
+		return nil
+	}
+
 	out := make(chan XMLRecord, ChanDepth())
 	if out == nil {
 		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create fetcher channel\n")
 		os.Exit(1)
 	}
 
-	if zipp {
-		sfx += ".gz"
-	}
-
-	fetchRecord := func(file string, buf bytes.Buffer) string {
-
-		var arry [132]rune
-		trie := MakeArchiveTrie(file, arry)
-
-		if file == "" || trie == "" {
-			return ""
-		}
-
-		fpath := path.Join(stash, trie, file+sfx)
-		if fpath == "" {
-			return ""
-		}
-
-		iszip := zipp
-
-		inFile, err := os.Open(fpath)
-
-		// if failed to find ".xml" or ".e2x" file, try ".xml.gz" or ".e2x.gz" without requiring -gzip
-		if err != nil && os.IsNotExist(err) && !zipp {
-			iszip = true
-			fpath := path.Join(stash, trie, file+sfx+".gz")
-			if fpath == "" {
-				return ""
-			}
-			inFile, err = os.Open(fpath)
-		}
-		if err != nil {
-			msg := err.Error()
-			if !strings.HasSuffix(msg, "no such file or directory") && !strings.HasSuffix(msg, "cannot find the path specified.") {
-				fmt.Fprintf(os.Stderr, "%s\n", msg)
-			}
-			return ""
-		}
-
-		defer inFile.Close()
-
-		brd := bufio.NewReader(inFile)
-
-		if iszip {
-
-			zpr, err := gzip.NewReader(brd)
-
-			defer zpr.Close()
-
-			if err == nil {
-				// copy and decompress cached file contents
-				buf.ReadFrom(zpr)
-			}
-
-		} else {
-
-			// copy cached file contents
-			buf.ReadFrom(brd)
-		}
-
-		str := buf.String()
-
-		return str
-	}
-
 	// xmlFetcher reads XML from file
 	xmlFetcher := func(wg *sync.WaitGroup, inp <-chan XMLRecord, out chan<- XMLRecord) {
 
@@ -591,11 +922,30 @@ func CreateFetchers(stash, sfx string, zipp bool, inp <-chan XMLRecord) <-chan X
 
 			buf.Reset()
 
-			str := fetchRecord(ext.Text, buf)
+			str := fetchOneXMLRecord(ext.Text, stsh, pfx, sfx, zipp, buf)
+
+			// trim header now included in archive XML files
+			if db == "" || db == "pubmed" {
+				if str != "" {
+					pos := strings.Index(str, "<PubmedArticle>")
+					if pos > 0 {
+						// remove leading xml and DOCTYPE lines
+						str = str[pos:]
+					}
+				}
+			} else if db == "pmc" {
+				if str != "" {
+					pos := strings.Index(str, "<PMCExtract>")
+					if pos > 0 {
+						// remove any leading xml and DOCTYPE lines
+						str = str[pos:]
+					}
+				}
+			}
 
 			runtime.Gosched()
 
-			out <- XMLRecord{Index: ext.Index, Text: str}
+			out <- XMLRecord{Index: ext.Index, Ident: ext.Ident, Text: str}
 		}
 	}
 
@@ -618,9 +968,9 @@ func CreateFetchers(stash, sfx string, zipp bool, inp <-chan XMLRecord) <-chan X
 
 // CreateCacheStreamers returns compressed records from archive, multithreaded for speed,
 // could be used for sending records over network to be decompressed later by client
-func CreateCacheStreamers(stash, sfx string, inp <-chan XMLRecord) <-chan XMLRecord {
+func CreateCacheStreamers(stsh, pfx, sfx string, inp <-chan XMLRecord) <-chan XMLRecord {
 
-	if inp == nil {
+	if inp == nil || stsh == "" {
 		return nil
 	}
 
@@ -629,19 +979,21 @@ func CreateCacheStreamers(stash, sfx string, inp <-chan XMLRecord) <-chan XMLRec
 		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create streamer channel\n")
 		os.Exit(1)
 	}
-
 	sfx += ".gz"
 
-	getRecord := func(file string, buf bytes.Buffer) []byte {
+	getRecord := func(id string, buf bytes.Buffer) []byte {
 
-		var arry [132]rune
-		trie := MakeArchiveTrie(file, arry)
-
-		if file == "" || trie == "" {
+		if id == "" {
 			return nil
 		}
 
-		fpath := path.Join(stash, trie, file+sfx)
+		dir, file := ArchiveTrie(id)
+
+		if dir == "" || file == "" {
+			return nil
+		}
+
+		fpath := filepath.Join(stsh, dir, pfx+file+sfx)
 		if fpath == "" {
 			return nil
 		}
@@ -684,7 +1036,12 @@ func CreateCacheStreamers(stash, sfx string, inp <-chan XMLRecord) <-chan XMLRec
 
 			runtime.Gosched()
 
-			out <- XMLRecord{Index: ext.Index, Data: data}
+			// skip past first gzip packet with xml and DOCTYPE
+			if len(data) > XMLDoctypeGzipLen {
+				data = data[XMLDoctypeGzipLen:]
+			}
+
+			out <- XMLRecord{Index: ext.Index, Ident: ext.Ident, Data: data}
 		}
 	}
 
