@@ -35,490 +35,110 @@ import (
 	"encoding/binary"
 	"fmt"
 	"github.com/surgebase/porter2"
+	"html"
 	"io"
 	"os"
-	"path"
+	"path/filepath"
 	"regexp"
 	"runtime"
 	"sort"
 	"strconv"
 	"strings"
+	"sync"
 	"unicode"
 )
 
-// Master points to a term and to its postings data
-type Master struct {
-	TermOffset int32
-	PostOffset int32
+type alias struct {
+	table    map[string]string
+	lock     sync.Mutex
+	fpath    string
+	isLoaded bool
 }
 
-// Arrays contains postings lists and word offsets
-type Arrays struct {
-	Data []int32
-	Ofst [][]int16
-	Dist int
-}
+// loadAliasTable should be called within a lock on the alias.lock mutex
+func (a *alias) loadAliasTable(reverse, commas bool) {
 
-func commonOpenFile(dpath, fname string) (*os.File, int64) {
-
-	fpath := path.Join(dpath, fname)
-	if fpath == "" {
-		return nil, 0
+	if a == nil || a.fpath == "" {
+		return
 	}
 
-	inFile, err := os.Open(fpath)
-	if err != nil && os.IsNotExist(err) {
-		return nil, 0
-	}
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil, 0
+	if a.isLoaded {
+		return
 	}
 
-	fi, err := inFile.Stat()
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil, 0
-	}
-
-	size := fi.Size()
-
-	return inFile, size
-}
-
-func readMasterIndex(dpath, key, field string) []Master {
-
-	inFile, size := commonOpenFile(dpath, key+"."+field+".mst")
-	if inFile == nil {
-		return nil
-	}
-
-	defer inFile.Close()
-
-	data := make([]Master, size/8)
-	if data == nil || len(data) < 1 {
-		return nil
-	}
-
-	err := binary.Read(inFile, binary.LittleEndian, &data)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	return data
-}
-
-func readTermList(dpath, key, field string) []byte {
-
-	inFile, size := commonOpenFile(dpath, key+"."+field+".trm")
-	if inFile == nil {
-		return nil
-	}
-
-	defer inFile.Close()
-
-	data := make([]byte, size)
-	if data == nil || len(data) < 1 {
-		return nil
-	}
-
-	err := binary.Read(inFile, binary.LittleEndian, &data)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	return data
-}
-
-func readPostingData(dpath, key, field string, offset int32, size int32) []int32 {
-
-	inFile, _ := commonOpenFile(dpath, key+"."+field+".pst")
-	if inFile == nil {
-		return nil
-	}
-
-	defer inFile.Close()
-
-	data := make([]int32, size/4)
-	if data == nil || len(data) < 1 {
-		return nil
-	}
-
-	_, err := inFile.Seek(int64(offset), io.SeekStart)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	err = binary.Read(inFile, binary.LittleEndian, data)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	return data
-}
-
-func readPositionIndex(dpath, key, field string, offset int32, size int32) []int32 {
-
-	inFile, _ := commonOpenFile(dpath, key+"."+field+".uqi")
-	if inFile == nil {
-		return nil
-	}
-
-	defer inFile.Close()
-
-	data := make([]int32, size/4)
-	if data == nil || len(data) < 1 {
-		return nil
-	}
-
-	_, err := inFile.Seek(int64(offset), io.SeekStart)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	err = binary.Read(inFile, binary.LittleEndian, data)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	return data
-}
-
-func readOffsetData(dpath, key, field string, offset int32, size int32) []int16 {
-
-	inFile, _ := commonOpenFile(dpath, key+"."+field+".ofs")
-	if inFile == nil {
-		return nil
-	}
-
-	defer inFile.Close()
-
-	data := make([]int16, size/2)
-	if data == nil || len(data) < 1 {
-		return nil
-	}
-
-	_, err := inFile.Seek(int64(offset), io.SeekStart)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	err = binary.Read(inFile, binary.LittleEndian, data)
-	if err != nil {
-		fmt.Fprintf(os.Stderr, "%s\n", err.Error())
-		return nil
-	}
-
-	return data
-}
-
-func readMasterIndexFuture(dpath, key, field string) <-chan []Master {
-
-	out := make(chan []Master, ChanDepth())
-	if out == nil {
-		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create master index channel\n")
-		os.Exit(1)
-	}
-
-	// masterIndexFuture asynchronously gets the master file and sends results through channel
-	masterIndexFuture := func(dpath, key, field string, out chan<- []Master) {
-
-		data := readMasterIndex(dpath, key, field)
-
-		out <- data
-
-		close(out)
-	}
-
-	// launch single future goroutine
-	go masterIndexFuture(dpath, key, field, out)
-
-	return out
-}
-
-func readTermListFuture(dpath, key, field string) <-chan []byte {
-
-	out := make(chan []byte, ChanDepth())
-	if out == nil {
-		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create term list channel\n")
-		os.Exit(1)
-	}
-
-	// termListFuture asynchronously gets posting IDs and sends results through channel
-	termListFuture := func(dpath, key, field string, out chan<- []byte) {
-
-		data := readTermList(dpath, key, field)
-
-		out <- data
-
-		close(out)
-	}
-
-	// launch single future goroutine
-	go termListFuture(dpath, key, field, out)
-
-	return out
-}
-
-func getPostingIDs(prom, term, field string, simple bool) ([]int32, [][]int16) {
-
-	var (
-		arry [516]rune
-	)
-
-	dpath, key := PostingPath(prom, field, term, arry)
-	if dpath == "" {
-		return nil, nil
-	}
-
-	// schedule asynchronous fetching
-	mi := readMasterIndexFuture(dpath, key, field)
-
-	tl := readTermListFuture(dpath, key, field)
-
-	// fetch master index and term list
-	indx := <-mi
-
-	trms := <-tl
-
-	if indx == nil || len(indx) < 1 {
-		return nil, nil
-	}
-
-	if trms == nil || len(trms) < 1 {
-		return nil, nil
-	}
-
-	// master index is padded with phantom term and postings position
-	numTerms := len(indx) - 1
-
-	strs := make([]string, numTerms)
-	if strs == nil || len(strs) < 1 {
-		return nil, nil
-	}
-
-	retlength := int32(len("\n"))
-
-	// populate array of strings from term list
-	for i, j := 0, 1; i < numTerms; i++ {
-		from := indx[i].TermOffset
-		to := indx[j].TermOffset - retlength
-		j++
-		txt := string(trms[from:to])
-		strs[i] = txt
-	}
-
-	// change protecting underscore to space
-	term = strings.Replace(term, "_", " ", -1)
-
-	// if term ends with dollar sign, use porter2 stemming, then add asterisk
-	if strings.HasSuffix(term, "$") && term != "$" {
-		term = strings.TrimSuffix(term, "$")
-		term = porter2.Stem(term)
-		term += "*"
-	}
-
-	isWildCard := false
-	if strings.HasSuffix(term, "*") && term != "*" {
-		tlen := len(term)
-		isWildCard = true
-		term = strings.TrimSuffix(term, "*")
-		pdlen := len(PostingDir(term))
-		if tlen < pdlen {
-			fmt.Fprintf(os.Stderr, "Wildcard term '%s' must be at least %d characters long - ignoring this word\n", term, pdlen)
-			return nil, nil
-		}
-	}
-
-	// binary search in term list
-	L, R := 0, numTerms-1
-	for L < R {
-		mid := (L + R) / 2
-		if strs[mid] < term {
-			L = mid + 1
-		} else {
-			R = mid
-		}
-	}
-
-	// wild card search scans term lists, fuses adjacent postings lists
-	if isWildCard {
-		if R < numTerms && strings.HasPrefix(strs[R], term) {
-			offset := indx[R].PostOffset
-			for R < numTerms && strings.HasPrefix(strs[R], term) {
-				R++
+	file, ferr := os.Open(a.fpath)
+
+	if file != nil && ferr == nil {
+		scanner := bufio.NewScanner(file)
+		for scanner.Scan() {
+			str := scanner.Text()
+			if str == "" {
+				continue
 			}
-			size := indx[R].PostOffset - offset
-
-			// read relevant postings list section
-			data := readPostingData(dpath, key, field, offset, size)
-			if data == nil || len(data) < 1 {
-				return nil, nil
+			cols := strings.SplitN(str, "\t", 2)
+			if len(cols) != 2 {
+				continue
 			}
 
-			if simple {
-
-				merged := make(map[int32]bool)
-
-				// combine all postings in term range
-				for _, val := range data {
-					merged[val] = true
+			cleanTerm := func(str string) string {
+				str = CleanupQuery(str, false, true)
+				parts := strings.FieldsFunc(str, func(c rune) bool {
+					return (!unicode.IsLetter(c) && !unicode.IsDigit(c) && c != ',') || c > 127
+				})
+				str = strings.Join(parts, " ")
+				if commas {
+					str = strings.Replace(str, ",", " ", -1)
 				}
-
-				fused := make([]int32, len(merged))
-
-				// convert map to slice
-				i := 0
-				for num := range merged {
-					fused[i] = num
-					i++
-				}
-
-				sort.Slice(fused, func(i, j int) bool { return fused[i] < fused[j] })
-
-				return fused, nil
+				str = strings.ToLower(str)
+				str = strings.TrimSpace(str)
+				str = CompressRunsOfSpaces(str)
+				return str
 			}
 
-			// read relevant word position section, includes phantom offset at end
-			uqis := readPositionIndex(dpath, key, field, offset, size+4)
-			if uqis == nil {
-				return nil, nil
+			one := cleanTerm(cols[0])
+			two := cleanTerm(cols[1])
+			if reverse {
+				// fmt.Fprintf(os.Stderr, "R '%s' -> '%s'\n", two, one)
+				a.table[two] = one
+			} else {
+				// fmt.Fprintf(os.Stderr, "F '%s' -> '%s'\n", one, two)
+				a.table[one] = two
 			}
-			ulen := len(uqis)
-			if ulen < 1 {
-				return nil, nil
-			}
-
-			from := uqis[0]
-			to := uqis[ulen-1]
-
-			// read offset section
-			ofst := readOffsetData(dpath, key, field, from, to-from)
-			if ofst == nil {
-				return nil, nil
-			}
-
-			combo := make(map[int32][]int16)
-
-			addPositions := func(uid int32, pos int16) {
-
-				arrs, ok := combo[uid]
-				if !ok {
-					arrs = make([]int16, 0, 1)
-				}
-				arrs = append(arrs, pos)
-				combo[uid] = arrs
-			}
-
-			// populate array of positions per UID
-			for i, j, k := 0, 1, int32(0); i < ulen-1; i++ {
-				uid := data[i]
-				num := (uqis[j] - uqis[i]) / 2
-				j++
-				for q := k; q < k+num; q++ {
-					addPositions(uid, ofst[q])
-				}
-				k += num
-			}
-
-			fused := make([]int32, len(combo))
-
-			// convert map to slice
-			i := 0
-			for num := range combo {
-				fused[i] = num
-				i++
-			}
-
-			sort.Slice(fused, func(i, j int) bool { return fused[i] < fused[j] })
-
-			// make array of int16 arrays, populate for each UID
-			arrs := make([][]int16, ulen-1)
-			if arrs == nil {
-				return nil, nil
-			}
-
-			for j, uid := range fused {
-				posn := combo[uid]
-
-				if len(posn) > 1 {
-					sort.Slice(posn, func(i, j int) bool { return posn[i] < posn[j] })
-				}
-
-				arrs[j] = posn
-			}
-
-			return fused, arrs
 		}
-
-		return nil, nil
 	}
 
-	// regular search requires exact match from binary search
-	if R < numTerms && strs[R] == term {
+	file.Close()
 
-		offset := indx[R].PostOffset
-		size := indx[R+1].PostOffset - offset
-
-		// read relevant postings list section
-		data := readPostingData(dpath, key, field, offset, size)
-		if data == nil || len(data) < 1 {
-			return nil, nil
-		}
-
-		if simple {
-			return data, nil
-		}
-
-		// read relevant word position section, includes phantom offset at end
-		uqis := readPositionIndex(dpath, key, field, offset, size+4)
-		if uqis == nil {
-			return nil, nil
-		}
-		ulen := len(uqis)
-		if ulen < 1 {
-			return nil, nil
-		}
-
-		from := uqis[0]
-		to := uqis[ulen-1]
-
-		// read offset section
-		ofst := readOffsetData(dpath, key, field, from, to-from)
-		if ofst == nil {
-			return nil, nil
-		}
-
-		// make array of int16 arrays, populate for each UID
-		arrs := make([][]int16, ulen)
-		if arrs == nil || len(arrs) < 1 {
-			return nil, nil
-		}
-
-		// populate array of positions per UID
-		for i, j, k := 0, 1, int32(0); i < ulen-1; i++ {
-			num := (uqis[j] - uqis[i]) / 2
-			j++
-			arrs[i] = ofst[k : k+num]
-			k += num
-		}
-
-		return data, arrs
-	}
-
-	return nil, nil
+	// set even if loading failed to prevent multiple attempts
+	a.isLoaded = true
 }
 
+var journalAliases = map[string]string{
+	"pnas":                  "proc natl acad sci u s a",
+	"journal of immunology": "journal of immunology baltimore md 1950",
+	"biorxiv":               "biorxiv the preprint server for biology",
+	"biorxivorg":            "biorxiv the preprint server for biology",
+}
+
+var ptypAliases = map[string]string{
+	"clinical trial phase 1":     "clinical trial phase i",
+	"clinical trial phase 2":     "clinical trial phase ii",
+	"clinical trial phase 3":     "clinical trial phase iii",
+	"clinical trial phase 4":     "clinical trial phase iv",
+	"clinical trial phase one":   "clinical trial phase i",
+	"clinical trial phase two":   "clinical trial phase ii",
+	"clinical trial phase three": "clinical trial phase iii",
+	"clinical trial phase four":  "clinical trial phase iv",
+}
+
+var (
+	meshName alias
+	meshTree alias
+)
+
 func printTermCount(base, term, field string) int {
 
-	data, _ := getPostingIDs(base, term, field, true)
+	data, _ := getPostingIDs(base, term, field, true, false)
 	size := len(data)
 	fmt.Fprintf(os.Stdout, "%d\t%s\n", size, term)
 
@@ -539,8 +159,7 @@ func printTermCounts(base, term, field string) int {
 		os.Exit(1)
 	}
 
-	var arry [516]rune
-	dpath, key := PostingPath(base, field, term, arry)
+	dpath, key := PostingPath(base, field, term, false)
 	if dpath == "" {
 		return 0
 	}
@@ -614,7 +233,7 @@ func printTermCounts(base, term, field string) int {
 
 func printTermPositions(base, term, field string) int {
 
-	data, ofst := getPostingIDs(base, term, field, false)
+	data, ofst := getPostingIDs(base, term, field, false, false)
 	size := len(data)
 	fmt.Fprintf(os.Stdout, "\n%d\t%s\n\n", size, term)
 
@@ -632,344 +251,9 @@ func printTermPositions(base, term, field string) int {
 	return size
 }
 
-// BOOLEAN OPERATIONS FOR POSTINGS LISTS
-
-func extendPositionalIDs(N []int32, np [][]int16, M []int32, mp [][]int16, delta int, proc func(pn, pm []int16, dlt int16) []int16) ([]int32, [][]int16) {
-
-	if proc == nil {
-		return nil, nil
-	}
-
-	if N == nil || len(N) < 1 || np == nil || len(np) < 1 {
-		return M, mp
-	}
-	if M == nil || len(M) < 1 || mp == nil || len(mp) < 1 {
-		return N, np
-	}
-
-	n, m := len(N), len(M)
-
-	// order matters when extending phrase or testing proximity, do not swap lists based on size
-
-	sz := n
-	if sz > m {
-		sz = m
-	}
-
-	if sz < 1 {
-		return N, np
-	}
-
-	res := make([]int32, sz)
-	ofs := make([][]int16, sz)
-
-	if res == nil || len(res) < 1 || ofs == nil || len(ofs) < 1 {
-		return nil, nil
-	}
-
-	i, j, k := 0, 0, 0
-
-	// use local variables for speed
-	en, em := N[i], M[j]
-
-	for {
-		// do inequality tests first
-		if en < em {
-			i++
-			if i == n {
-				break
-			}
-			en = N[i]
-		} else if en > em {
-			j++
-			if j == m {
-				break
-			}
-			em = M[j]
-		} else {
-			// specific callbacks test position arrays to match terms by adjacency or phrases by proximity
-			adj := proc(np[i], mp[j], int16(delta))
-			if adj != nil && len(adj) > 0 {
-				res[k] = en
-				ofs[k] = adj
-				k++
-			}
-			i++
-			j++
-			if i == n || j == m {
-				break
-			}
-			en = N[i]
-			em = M[j]
-		}
-	}
-
-	// truncate output arrays to actual size of intersection
-	res = res[:k]
-	ofs = ofs[:k]
-
-	return res, ofs
-}
-
-func intersectIDs(N, M []int32) []int32 {
-
-	if N == nil {
-		return M
-	}
-	if M == nil {
-		return N
-	}
-
-	n, m := len(N), len(M)
-
-	// swap to make M the smaller list
-	if n < m {
-		N, M = M, N
-		n, m = m, n
-	}
-
-	if m < 1 {
-		return N
-	}
-
-	res := make([]int32, m)
-
-	i, j, k := 0, 0, 0
-
-	// use local variables for speed
-	en, em := N[i], M[j]
-
-	for {
-		// do inequality tests first
-		if en < em {
-			// index to larger list most likely to be advanced
-			i++
-			if i == n {
-				break
-			}
-			en = N[i]
-		} else if en > em {
-			j++
-			if j == m {
-				break
-			}
-			em = M[j]
-		} else {
-			// equality (intersection match) least likely
-			res[k] = en
-			k++
-			i++
-			j++
-			if i == n || j == m {
-				break
-			}
-			en = N[i]
-			em = M[j]
-		}
-	}
-
-	// truncate output array to actual size of intersection
-	res = res[:k]
-
-	return res
-}
-
-// if m * log(n) < m + n, binary search has fewer comparisons, but processor memory caches make linear algorithm faster
-/*
-func intersectBinary(N, M []int32) []int32 {
-
-	if N == nil {
-		return M
-	}
-	if M == nil {
-		return N
-	}
-
-	n, m := len(N), len(M)
-
-	// swap to make M the smaller list
-	if n < m {
-		N, M = M, N
-		n, m = m, n
-	}
-
-	if m < 1 {
-		return N
-	}
-
-	k := 0
-
-	res := make([]int32, m)
-
-	for _, uid := range M {
-		// inline binary search is faster than sort.Search
-		L, R := 0, n-1
-		for L < R {
-			mid := (L + R) / 2
-			if N[mid] < uid {
-				L = mid + 1
-			} else {
-				R = mid
-			}
-		}
-		// R := sort.Search(len(N), func(i int) bool { return N[i] >= uid })
-		if R < n && N[R] == uid {
-			res[k] = uid
-			k++
-			// remove leading part of N for slight speed gain
-			N = N[R:]
-			n = len(N)
-		}
-	}
-
-	res = res[:k]
-
-	return res
-}
-*/
-
-func combineIDs(N, M []int32) []int32 {
-
-	if N == nil {
-		return M
-	}
-	if M == nil {
-		return N
-	}
-
-	n, m := len(N), len(M)
-
-	// swap to make M the smaller list
-	if n < m {
-		N, M = M, N
-		n, m = m, n
-	}
-
-	if m < 1 {
-		return N
-	}
-
-	i, j, k := 0, 0, 0
-
-	res := make([]int32, n+m)
-
-	for i < n && j < m {
-		if N[i] < M[j] {
-			res[k] = N[i]
-			k++
-			i++
-		} else if N[i] > M[j] {
-			res[k] = M[j]
-			k++
-			j++
-		} else {
-			res[k] = N[i]
-			k++
-			i++
-			j++
-		}
-	}
-	for i < n {
-		res[k] = N[i]
-		k++
-		i++
-	}
-	for j < m {
-		res[k] = M[j]
-		k++
-		j++
-	}
-
-	res = res[:k]
-
-	return res
-}
-
-func excludeIDs(N, M []int32) []int32 {
-
-	if N == nil {
-		return nil
-	}
-	if M == nil {
-		return N
-	}
-
-	n, m := len(N), len(M)
-
-	if m < 1 {
-		return N
-	}
-
-	res := make([]int32, n)
-
-	i, j, k := 0, 0, 0
-
-	// use local variables for speed
-	en, em := N[i], M[j]
-
-	for {
-		// do inequality tests first
-		if en < em {
-			// item is not excluded
-			res[k] = en
-			k++
-			i++
-			if i == n {
-				break
-			}
-			en = N[i]
-		} else if en > em {
-			// advance second list pointer
-			j++
-			if j == m {
-				break
-			}
-			em = M[j]
-		} else {
-			// exclude
-			i++
-			j++
-			if i == n || j == m {
-				break
-			}
-			en = N[i]
-			em = M[j]
-		}
-	}
-
-	// truncate output array to actual size of result
-	res = res[:k]
-
-	return res
-}
-
 // QUERY EVALUATION FUNCTION
 
-func postingIDsFuture(base, term, field string, dist int) <-chan Arrays {
-
-	out := make(chan Arrays, ChanDepth())
-	if out == nil {
-		fmt.Fprintf(os.Stderr, "\nERROR: Unable to create postings channel\n")
-		os.Exit(1)
-	}
-
-	// postingFuture asynchronously gets posting IDs and sends results through channel
-	postingFuture := func(base, term, field string, dist int, out chan<- Arrays) {
-
-		data, ofst := getPostingIDs(base, term, field, false)
-
-		out <- Arrays{Data: data, Ofst: ofst, Dist: dist}
-
-		close(out)
-	}
-
-	// launch single future goroutine
-	go postingFuture(base, term, field, dist, out)
-
-	return out
-}
-
-func evaluateQuery(base string, clauses []string, noStdout bool) (int, []int32) {
+func evaluateQuery(base, dbase, phrase string, clauses []string, noStdout, isLink bool) (int, []int32) {
 
 	if clauses == nil || clauses[0] == "" {
 		return 0, nil
@@ -1073,6 +357,9 @@ func evaluateQuery(base string, clauses []string, noStdout bool) (int, []int32)
 
 		// extract optional [FIELD] qualifier
 		field := "TIAB"
+		if dbase == "pmc" {
+			field = "TEXT"
+		}
 
 		if strings.HasSuffix(str, "]") {
 			pos := strings.Index(str, "[")
@@ -1086,7 +373,7 @@ func evaluateQuery(base string, clauses []string, noStdout bool) (int, []int32)
 			switch field {
 			case "NORM":
 				field = "TIAB"
-			case "STEM", "TIAB", "TITL":
+			case "STEM", "TIAB", "TITL", "ABST", "TEXT":
 			case "PIPE":
 				// esearch -db pubmed -query "complement system proteins [MESH]" -pub clinical |
 				// efetch -format uid | phrase-search -query "[PIPE] AND L [THME]"
@@ -1125,7 +412,7 @@ func evaluateQuery(base string, clauses []string, noStdout bool) (int, []int32)
 				return nil, nil, 0
 			}
 			term = strings.Replace(term, "_", " ", -1)
-			data, _ := getPostingIDs(base, term, field, true)
+			data, _ := getPostingIDs(base, term, field, true, isLink)
 			count++
 			return data, nil, 1
 		}
@@ -1147,7 +434,7 @@ func evaluateQuery(base string, clauses []string, noStdout bool) (int, []int32)
 				continue
 			}
 
-			fetch := postingIDsFuture(base, term, field, dist)
+			fetch := postingIDsFuture(base, term, field, dist, isLink)
 
 			futures = append(futures, fetch)
 
@@ -1261,7 +548,7 @@ func evaluateQuery(base string, clauses []string, noStdout bool) (int, []int32)
 			fmt.Fprintf(os.Stderr, "\nERROR: Unexpected operator '%s' in expression\n", tkn)
 			os.Exit(1)
 		} else if tkn == "" {
-			fmt.Fprintf(os.Stderr, "\nERROR: Unexpected end of expression\n")
+			fmt.Fprintf(os.Stderr, "\nERROR: Unexpected end of expression in '%s'\n", phrase)
 			os.Exit(1)
 		} else {
 			// evaluate current phrase
@@ -1392,6 +679,8 @@ func prepareQuery(str string) string {
 		str = "stdin " + str
 	}
 
+	str = html.UnescapeString(str)
+
 	str = CleanupQuery(str, false, true)
 
 	str = strings.Replace(str, "~ ~", "~~", -1)
@@ -1484,6 +773,11 @@ func prepareQuery(str string) string {
 
 	str = strings.Replace(str, "-", " ", -1)
 
+	str = strings.Replace(str, "'", "", -1)
+
+	// allow links like pubmed_cited and pubmed_cites
+	str = strings.Replace(str, "[pubmed ", "[pubmed_", -1)
+
 	// break terms at punctuation, and at non-ASCII characters, allowing brackets for field names,
 	// along with Boolean control symbols, underscore for protected terms, asterisk to indicate
 	// truncation wildcard, tilde for maximum proximity, and plus sign for exactly one wildcard word
@@ -1553,7 +847,7 @@ func prepareExact(str, sfx string, deStop bool) string {
 
 	// break clauses at punctuation other than space or underscore, and at non-ASCII characters
 	clauses := strings.FieldsFunc(cleaned, func(c rune) bool {
-		return (!unicode.IsLetter(c) && !unicode.IsDigit(c)) && c != ' ' && c != '_' || c > 127
+		return (!unicode.IsLetter(c) && !unicode.IsDigit(c) && c != ' ' && c != '_') || c > 127
 	})
 
 	// space replaces plus sign to separate runs of unpunctuated words
@@ -1640,7 +934,7 @@ func processStopWords(str string, deStop bool) string {
 
 		stps := false
 		rlxd := false
-		if fld == "[TITL]" || fld == "[TIAB]" {
+		if fld == "[TITL]" || fld == "[TIAB]" || fld == "[ABST]" || fld == "[TEXT]" {
 			stps = true
 		} else if fld == "[STEM]" {
 			stps = true
@@ -1817,6 +1111,12 @@ func setFieldQualifiers(clauses []string, rlxd bool) []string {
 				os.Exit(1)
 			}
 
+			// allow year month day to look for unexpected annotation
+			if len(str) > 9 {
+				res = append(res, str+" [YEAR]")
+				continue
+			}
+
 			// check for year range
 			if len(str) == 9 && str[4] == ' ' && IsAllDigitsOrPeriod(str[:4]) && IsAllDigitsOrPeriod(str[5:]) {
 				start, err := strconv.Atoi(str[:4])
@@ -1850,6 +1150,8 @@ func setFieldQualifiers(clauses []string, rlxd bool) []string {
 			os.Exit(1)
 
 		} else if strings.HasSuffix(str, " [ANUM]") ||
+			strings.HasSuffix(str, " [INUM]") ||
+			strings.HasSuffix(str, " [FNUM]") ||
 			strings.HasSuffix(str, " [TLEN]") ||
 			strings.HasSuffix(str, " [TNUM]") {
 
@@ -1944,6 +1246,88 @@ func setFieldQualifiers(clauses []string, rlxd bool) []string {
 
 			fmt.Fprintf(os.Stderr, "\nERROR: Unable to recognize mesh code expression '%s'\n", str)
 			os.Exit(1)
+
+		} else if strings.HasSuffix(str, " [JOUR]") {
+
+			slen := len(str)
+			str = str[:slen-7]
+
+			// check hard-coded journal alias map (would be better to use map from Data/jourindx.txt)
+			alias, ok := journalAliases[str]
+			if ok {
+				res = append(res, alias+" [JOUR]")
+				continue
+			}
+
+			// no alias found, use as is
+			res = append(res, str+" [JOUR]")
+			continue
+
+		} else if strings.HasSuffix(str, " [PTYP]") {
+
+			slen := len(str)
+			str = str[:slen-7]
+
+			// convert clinical trial phase with arabic numeral or english word to roman numeral
+			alias, ok := ptypAliases[str]
+			if ok {
+				res = append(res, alias+" [PTYP]")
+				continue
+			}
+
+			// no alias found, use as is
+			res = append(res, str+" [PTYP]")
+			continue
+
+		} else if strings.HasSuffix(str, " [MESH]") {
+
+			slen := len(str)
+			str = str[:slen-7]
+
+			// load mesh tables within mutexes
+			meshName.lock.Lock()
+			if !meshName.isLoaded {
+				meshName.loadAliasTable(true, true)
+			}
+			meshName.lock.Unlock()
+
+			meshTree.lock.Lock()
+			if !meshTree.isLoaded {
+				meshTree.loadAliasTable(false, false)
+			}
+			meshTree.lock.Unlock()
+
+			// check mesh alias tables
+			if meshName.isLoaded && meshTree.isLoaded {
+				code, ok := meshName.table[str]
+				if ok {
+					cluster, ok := meshTree.table[code]
+					if ok {
+						if strings.Index(cluster, ",") < 0 {
+							res = append(res, cluster+"* [TREE]")
+							continue
+						}
+						trees := strings.Split(cluster, ",")
+						// expand multiple trees in OR group
+						pfx := "("
+						sfx := ")"
+						for _, tr := range trees {
+							res = append(res, pfx)
+							pfx = "|"
+							tr = strings.TrimSpace(tr)
+							res = append(res, tr+"* [TREE]")
+						}
+						res = append(res, sfx)
+						continue
+					} else {
+						res = append(res, code+" [CODE]")
+						continue
+					}
+				}
+			}
+
+			// skip if MeSH term not yet indexed in tree
+			continue
 		}
 
 		// remove leading and trailing plus signs and spaces
@@ -1964,16 +1348,31 @@ func setFieldQualifiers(clauses []string, rlxd bool) []string {
 // SEARCH TERM LISTS FOR PHRASES OR NORMALIZED TERMS, OR MATCH BY PATTERN
 
 // ProcessSearch evaluates query, returns list of PMIDs to stdout
-func ProcessSearch(base, phrase string, xact, titl, rlxd, deStop bool) int {
+func ProcessSearch(base, dbase, phrase string, xact, titl, rlxd, isLink, deStop bool) int {
 
 	if phrase == "" {
 		return 0
 	}
 
+	if base == "" {
+		// obtain path from environment variable within rchive as a convenience
+		base = os.Getenv("EDIRECT_PUBMED_MASTER")
+		if base != "" {
+			if !strings.HasSuffix(base, "/") {
+				base += "/"
+			}
+			base += "Postings"
+		}
+	}
+
 	if titl {
 		phrase = prepareExact(phrase, "[titl]", deStop)
 	} else if xact {
-		phrase = prepareExact(phrase, "[tiab]", deStop)
+		if dbase == "pmc" {
+			phrase = prepareExact(phrase, "[text]", deStop)
+		} else {
+			phrase = prepareExact(phrase, "[tiab]", deStop)
+		}
 	} else {
 		phrase = prepareQuery(phrase)
 	}
@@ -1984,22 +1383,37 @@ func ProcessSearch(base, phrase string, xact, titl, rlxd, deStop bool) int {
 
 	clauses = setFieldQualifiers(clauses, rlxd)
 
-	count, _ := evaluateQuery(base, clauses, false)
+	count, _ := evaluateQuery(base, dbase, phrase, clauses, false, isLink)
 
 	return count
 }
 
 // ProcessQuery evaluates query, returns list of PMIDs in array
-func ProcessQuery(base, phrase string, xact, titl, rlxd, deStop bool) []int32 {
+func ProcessQuery(base, dbase, phrase string, xact, titl, rlxd, isLink, deStop bool) []int32 {
 
 	if phrase == "" {
 		return nil
 	}
 
+	if base == "" {
+		// obtain path from environment variable within rchive as a convenience
+		base = os.Getenv("EDIRECT_PUBMED_MASTER")
+		if base != "" {
+			if !strings.HasSuffix(base, "/") {
+				base += "/"
+			}
+			base += "Postings"
+		}
+	}
+
 	if titl {
 		phrase = prepareExact(phrase, "[titl]", deStop)
 	} else if xact {
-		phrase = prepareExact(phrase, "[tiab]", deStop)
+		if dbase == "pmc" {
+			phrase = prepareExact(phrase, "[text]", deStop)
+		} else {
+			phrase = prepareExact(phrase, "[tiab]", deStop)
+		}
 	} else {
 		phrase = prepareQuery(phrase)
 	}
@@ -2010,13 +1424,13 @@ func ProcessQuery(base, phrase string, xact, titl, rlxd, deStop bool) []int32 {
 
 	clauses = setFieldQualifiers(clauses, rlxd)
 
-	_, arry := evaluateQuery(base, clauses, true)
+	_, arry := evaluateQuery(base, dbase, phrase, clauses, true, isLink)
 
 	return arry
 }
 
 // ProcessMock shows individual steps in processing query for evaluation
-func ProcessMock(base, phrase string, xact, titl, rlxd, deStop bool) int {
+func ProcessMock(base, dbase, phrase string, xact, titl, rlxd, deStop bool) int {
 
 	if phrase == "" {
 		return 0
@@ -2029,7 +1443,11 @@ func ProcessMock(base, phrase string, xact, titl, rlxd, deStop bool) int {
 
 		fmt.Fprintf(os.Stdout, "prepareExact:\n\n%s\n\n", phrase)
 	} else if xact {
-		phrase = prepareExact(phrase, "[tiab]", deStop)
+		if dbase == "pmc" {
+			phrase = prepareExact(phrase, "[text]", deStop)
+		} else {
+			phrase = prepareExact(phrase, "[tiab]", deStop)
+		}
 
 		fmt.Fprintf(os.Stdout, "prepareExact:\n\n%s\n\n", phrase)
 	} else {
@@ -2062,12 +1480,23 @@ func ProcessMock(base, phrase string, xact, titl, rlxd, deStop bool) int {
 }
 
 // ProcessCount prints document count for each term, also supports terminal wildcards
-func ProcessCount(base, phrase string, plrl, psns, rlxd, deStop bool) int {
+func ProcessCount(base, dbase, phrase string, plrl, psns, rlxd, deStop bool) int {
 
 	if phrase == "" {
 		return 0
 	}
 
+	if base == "" {
+		// obtain path from environment variable within rchive as a convenience
+		base = os.Getenv("EDIRECT_PUBMED_MASTER")
+		if base != "" {
+			if !strings.HasSuffix(base, "/") {
+				base += "/"
+			}
+			base += "Postings"
+		}
+	}
+
 	phrase = prepareQuery(phrase)
 
 	phrase = processStopWords(phrase, deStop)
@@ -2116,6 +1545,9 @@ func ProcessCount(base, phrase string, plrl, psns, rlxd, deStop bool) int {
 	parseField := func(str string) (string, string) {
 
 		field := "TIAB"
+		if dbase == "pmc" {
+			field = "TEXT"
+		}
 
 		if strings.HasSuffix(str, "]") {
 			pos := strings.Index(str, "[")
@@ -2129,7 +1561,7 @@ func ProcessCount(base, phrase string, plrl, psns, rlxd, deStop bool) int {
 			switch field {
 			case "NORM":
 				field = "TIAB"
-			case "STEM", "TIAB", "TITL":
+			case "STEM", "TIAB", "TITL", "ABST", "TEXT":
 			case "PIPE":
 			default:
 				str = strings.Replace(str, " ", "_", -1)
@@ -2179,3 +1611,359 @@ func ProcessCount(base, phrase string, plrl, psns, rlxd, deStop bool) int {
 
 	return count
 }
+
+// TermCounts prints document counts for terms by subdirectory
+func TermCounts(dpath, key, field string) int {
+
+	if dpath == "" {
+		return 0
+	}
+
+	// schedule asynchronous fetching
+	mi := readMasterIndexFuture(dpath, key, field)
+
+	tl := readTermListFuture(dpath, key, field)
+
+	// fetch master index and term list
+	indx := <-mi
+
+	trms := <-tl
+
+	if indx == nil || len(indx) < 1 {
+		return 0
+	}
+
+	if trms == nil || len(trms) < 1 {
+		return 0
+	}
+
+	// master index is padded with phantom term and postings position
+	numTerms := len(indx) - 1
+
+	strs := make([]string, numTerms)
+	if strs == nil || len(strs) < 1 {
+		return 0
+	}
+
+	retlength := int32(len("\n"))
+
+	// populate array of strings from term list
+	for i, j := 0, 1; i < numTerms; i++ {
+		from := indx[i].TermOffset
+		to := indx[j].TermOffset - retlength
+		j++
+		txt := string(trms[from:to])
+		strs[i] = txt
+	}
+
+	count := 0
+
+	for R, str := range strs {
+		offset := indx[R].PostOffset
+		size := indx[R+1].PostOffset - offset
+		fmt.Fprintf(os.Stdout, "%d\t%s\n", size/4, str)
+		count++
+	}
+
+	return count
+}
+
+// ProcessLinks reads a list of PMIDs, merges resulting links
+func ProcessLinks(base, fld string) {
+
+	if fld == "" {
+		return
+	}
+
+	if base == "" {
+		// obtain path from environment variable as a convenience
+		base = os.Getenv("EDIRECT_PUBMED_MASTER")
+		if base != "" {
+			if !strings.HasSuffix(base, "/") {
+				base += "/"
+			}
+			base += "Postings"
+		}
+	}
+
+	// createLinkGrouper reads from UID reader and groups PMIDs under the same LinksTrie
+	createLinkGrouper := func(base, fld string, inp <-chan XMLRecord) <-chan []string {
+
+		if base == "" || fld == "" || inp == nil {
+			return nil
+		}
+
+		out := make(chan []string, ChanDepth())
+		if out == nil {
+			fmt.Fprintf(os.Stderr, "\nERROR: Unable to create link grouper channel\n")
+			os.Exit(1)
+		}
+
+		linkGrouper := func(base, fld string, inp <-chan XMLRecord, out chan<- []string) {
+
+			// report when grouper has no more records to process
+			defer close(out)
+
+			var arry []string
+
+			currPfx := ""
+
+			for ext := range inp {
+
+				uid := ext.Text
+				_, pfx := LinksTrie(uid, true)
+
+				if pfx != currPfx && currPfx != "" {
+
+					if arry != nil {
+						// send group of PMIDs with the same line trie down the channel
+						out <- arry
+					}
+
+					// empty the slice
+					arry = nil
+				}
+
+				arry = append(arry, uid)
+
+				currPfx = pfx
+			}
+
+			// send final results
+			if arry != nil {
+				// send group of PMIDs with the same line trie down the channel
+				out <- arry
+			}
+		}
+
+		// launch single link grouper goroutine
+		go linkGrouper(base, fld, inp, out)
+
+		return out
+	}
+
+	// mutex for link results
+	var llock sync.RWMutex
+
+	// map for combining link results
+	combinedLinks := make(map[int]bool)
+
+	createLinkMergers := func(prom, field string, inp <-chan []string) <-chan string {
+
+		if prom == "" || field == "" || inp == nil {
+			return nil
+		}
+
+		out := make(chan string, ChanDepth())
+		if out == nil {
+			fmt.Fprintf(os.Stderr, "\nERROR: Unable to create link merger channel\n")
+			os.Exit(1)
+		}
+
+		// linkMerge processes a set of terms from the same master index area
+		linkMerge := func(wg *sync.WaitGroup, prom, field string, inp <-chan []string, out chan<- string) {
+
+			// report when this matcher has no more records to process
+			defer wg.Done()
+
+			if inp == nil || out == nil {
+				return
+			}
+
+			for terms := range inp {
+
+				key := terms[0]
+
+				dir, ky := LinksTrie(key, true)
+				if dir == "" {
+					continue
+				}
+				dpath := filepath.Join(prom, field, dir)
+				if dpath == "" {
+					continue
+				}
+
+				// schedule asynchronous fetching
+				mi := readMasterIndexFuture(dpath, ky, field)
+
+				tl := readTermListFuture(dpath, ky, field)
+
+				// fetch master index and term list
+				indx := <-mi
+
+				trms := <-tl
+
+				if indx == nil || len(indx) < 1 {
+					continue
+				}
+
+				if trms == nil || len(trms) < 1 {
+					continue
+				}
+
+				// master index is padded with phantom term and postings position
+				numTerms := len(indx) - 1
+
+				strs := make([]string, numTerms)
+				if strs == nil || len(strs) < 1 {
+					continue
+				}
+
+				retlength := int32(len("\n"))
+
+				// populate array of strings from term list
+				for i, j := 0, 1; i < numTerms; i++ {
+					from := indx[i].TermOffset
+					to := indx[j].TermOffset - retlength
+					j++
+					txt := string(trms[from:to])
+					strs[i] = txt
+				}
+
+				postingsLoop := func(dpath, ky, field string) {
+
+					inFile, _ := commonOpenFile(dpath, ky+"."+field+".pst")
+					if inFile == nil {
+						return
+					}
+
+					defer inFile.Close()
+
+					for _, term := range terms {
+
+						term = PadNumericID(term)
+
+						// binary search in term list
+						L, R := 0, numTerms-1
+						for L < R {
+							mid := (L + R) / 2
+							if strs[mid] < term {
+								L = mid + 1
+							} else {
+								R = mid
+							}
+						}
+
+						linkLoop := func(offset, size int32) {
+
+							data := make([]int32, size/4)
+							if data == nil || len(data) < 1 {
+								return
+							}
+
+							_, err := inFile.Seek(int64(offset), io.SeekStart)
+							if err != nil {
+								fmt.Fprintf(os.Stderr, "%s\n", err.Error())
+								return
+							}
+
+							// read relevant postings list section
+							err = binary.Read(inFile, binary.LittleEndian, data)
+							if err != nil {
+								fmt.Fprintf(os.Stderr, "%s\n", err.Error())
+								return
+							}
+
+							if data == nil || len(data) < 1 {
+								return
+							}
+
+							llock.Lock()
+
+							for _, uid := range data {
+								combinedLinks[int(uid)] = true
+							}
+
+							llock.Unlock()
+						}
+
+						// regular search requires exact match from binary search
+						if R < numTerms && strs[R] == term {
+
+							offset := indx[R].PostOffset
+							size := indx[R+1].PostOffset - offset
+
+							linkLoop(offset, size)
+						}
+					}
+				}
+
+				postingsLoop(dpath, ky, field)
+
+				out <- ky
+			}
+		}
+
+		var wg sync.WaitGroup
+
+		// launch multiple link merger goroutines
+		for i := 0; i < NumServe(); i++ {
+			wg.Add(1)
+			go linkMerge(&wg, prom, field, inp, out)
+		}
+
+		// launch separate anonymous goroutine to wait until all mergers are done
+		go func() {
+			wg.Wait()
+			close(out)
+		}()
+
+		return out
+	}
+
+	// read text PMIDs from stdin
+	uidq := CreateUIDReader(os.Stdin)
+
+	grpq := createLinkGrouper(base, fld, uidq)
+
+	lnkq := createLinkMergers(base, fld, grpq)
+
+	// drain channel
+	for range lnkq {
+	}
+
+	// sort id keys in alphabetical order
+	var keys []int
+	for ky := range combinedLinks {
+		keys = append(keys, ky)
+	}
+
+	sort.Slice(keys, func(i, j int) bool { return keys[i] < keys[j] })
+
+	// use buffers to speed up PMID printing
+	var buffer strings.Builder
+
+	wrtr := bufio.NewWriter(os.Stdout)
+
+	for _, uid := range keys {
+		pmid := strconv.Itoa(uid)
+		buffer.WriteString(pmid)
+		buffer.WriteString("\n")
+	}
+
+	txt := buffer.String()
+	if txt != "" {
+		// print buffer
+		wrtr.WriteString(txt[:])
+	}
+
+	wrtr.Flush()
+
+	runtime.Gosched()
+}
+
+// initialize empty journal and MeSH maps before non-init functions are called
+func init() {
+
+	meshName.table = make(map[string]string)
+	meshTree.table = make(map[string]string)
+
+	nv := os.Getenv("EDIRECT_PUBMED_MASTER")
+	if nv != "" {
+		if !strings.HasSuffix(nv, "/") {
+			nv += "/"
+		}
+		meshName.fpath = filepath.Join(nv, "Data", "meshname.txt")
+		meshTree.fpath = filepath.Join(nv, "Data", "meshtree.txt")
+	}
+}
